/* =======================================
 * Projekt: Fliegender Roboter / Zeppelin
 * =======================================
 * Code für den Arduino
 * 
 * Aufgaben:
 * misst und übermittelt die Sensordaten an den Computer (Processing) und
 * empfängt die Steuerungsdaten für die Aktuatoren (Motoren) und steuert diese an
 * 
 */

// #include <SoftwareSerial.h>
#include <Servo.h>

/*
 * Falls SoftwareSerial verwendet werden soll,
 * müssen folgende Zeilen auskommentiert sein:
 * 
 * #define serialObject Serial
 * 
 * 
 * und dürfen folgende Zeilen nicht auskommentiert sein:
 * 
 * #include <SoftwareSerial.h>
 * const byte ARDUINO_RX_PIN = 5;
 * const byte ARDUINO_TX_PIN = 6;
 * SoftwareSerial serialObject(ARDUINO_RX_PIN, ARDUINO_TX_PIN);
 * 
 * if (serialObject.available() > 0) { // Aufruf erfolgt bei SoftwareSerial nicht automatisch
 *   serialEvent();
 * }
 * 
 */

#define serialObject Serial // für SoftwareSerial auskommentieren

Servo servoFinUpDown;
Servo servoFinLeftRight;
Servo servoESC;

// Pinbelegung:

// Digital Pins der Ultraschallsensoren:
const byte ULTRA_SONIC_RIGHT_PIN = 2;
const byte ULTRA_SONIC_LEFT_PIN = 3;

//const byte ARDUINO_RX_PIN = 5; // für SoftwareSerial Kommentierung aufheben
//const byte ARDUINO_TX_PIN = 6; // für SoftwareSerial Kommentierung aufheben
const byte SERVO_FIN_UP_DOWN_PIN = 9;
const byte SERVO_FIN_LEFT_RIGHT_PIN = 10;
const byte SERVO_ESC_PIN = 11;
const byte BATTERY_VOLTAGE_PIN = A0;
const byte PT_RIGHT_PIN = A5;
const byte PT_LEFT_PIN = A6;
const byte PT_DOWN_PIN = A7;

// Keys für die serielle Kommunikation im Format {key:value}[ZEILENUMBRUCH]
//
const byte TIMESTAMP = 0; // aktueller Zeitstempel
const byte LOG = 250;     // für andere Nachrichten
const byte ECHO = 254;    // für Echo-Tests
const byte ERROR = 255;   // Für möglicherweise auftretende Fehler (falscher Datentyp, falscher Wertebereich, etc.)

// Codes der Aktuatoren:
const byte FIN_UP_DOWN = 1;
const byte FIN_LEFT_RIGHT = 2;
const byte MOTOR_DRIVE = 3;

// Codes der Sensoren:
const byte DISTANCE_LEFT = 4;
const byte DISTANCE_RIGHT = 5;
const byte PT_RIGHT = 6;
const byte PT_LEFT = 7;
const byte PT_DOWN = 8;

// Code des Akkus:
const byte BATTERY_VOLTAGE = 9;

// zum Speichern, ob der entsprechende Servo/Motor auf dem neuesten Stand ist:
bool finUpDownNeedsUpdate = false;
bool finLeftRightNeedsUpdate = false;
bool motorDriveNeedsUpdate = false;

// zum Speichern des aktuellen Wertes:
byte newFinUpDownPosition;
byte newFinLeftRightPosition;
byte newMotorDrivePosition;

//SoftwareSerial serialObject(ARDUINO_RX_PIN, ARDUINO_TX_PIN); // für SoftwareSerial Kommentierung aufheben

// Sendet ein Key-Value-Pair im Format {key:value}[ZEILENUMBRUCH] über die serielle Schnittstelle
void sendData(byte key, String value) {
  serialObject.print("{" + String(key) + ":" + value + "}\n");
}

// Wird am Ende des Loop-Durchlaufs aufgerufen, wenn seriell empfangene Daten verfügbar sind
// Sendet Fehlernachricht zurück, wenn Daten unvollständig sind
// Ruft bei entsprechenden Keys bestimmte Funktionen auf
void serialEventt() {
  sendData(LOG, String(serialObject.available()) + 'b');
  String data = serialObject.readStringUntil(10);
  if (!data.startsWith(String('{')) || !data.endsWith(String('}'))) {
    const String message = "Anfangs- oder Endtag fehlt. Daten unvollständig.";
    sendData(ERROR, message);
    return;
  } else {
    const int collonPos = data.indexOf(':');
    if (collonPos == -1) {
      const String message = "Doppelpunkt als Trenner kann nicht gefunden werden.";
      sendData(ERROR, message);
      return;
    } else {
      const byte key = data.substring(1, collonPos).toInt();
      const String value = data.substring(collonPos + 1, data.length() - 1);
      // serialObject.println(value); // zum Debuggen
      switch (key) {
        case TIMESTAMP:
          break;
        case FIN_UP_DOWN:
          newFinUpDownPosition = value.toInt();
          finUpDownNeedsUpdate = true;
          break;
        case FIN_LEFT_RIGHT:
          newFinLeftRightPosition = value.toInt();
          finLeftRightNeedsUpdate = true;
          break;
        case MOTOR_DRIVE:
          newMotorDrivePosition = value.toInt();
          motorDriveNeedsUpdate = true;
          break;
        case ECHO:
          sendTimestamp();
          sendData(key, value);
          break;
        case ERROR:
          break;
        default:
          sendData(ERROR, "Falscher Key");
          break;
      }
    }
  }
}

// bewegt das Höhenruder
bool writeFinUpDownPosition(byte value) {
  static unsigned long lastCall;
  if (millis() - lastCall > 100) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 100ms liegt.
    lastCall = millis();
    servoFinUpDown.write(value);
    return true;
  } else {
    return false;
  }
}

// bewegt das Seitenruder
bool writeFinLeftRightPosition(byte value) {
  static unsigned long lastCall;
  if (millis() - lastCall > 100) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 100ms liegt.
    lastCall = millis();
    servoFinLeftRight.write(value);
    return true;
  } else {
    return false;
  }
}

// legt die Motorgeschwindigkeit fest
bool writeMotorDrive(byte value) {
  static unsigned long lastCall;
  if (millis() - lastCall > 100) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 100ms liegt.
    lastCall = millis();
    servoESC.write(value);
    return true;
  } else {
    return false;
  }
}

// misst und sendet die Daten des linken Ulltraschallsensors
bool sendDistanceLeft() {
  static unsigned long lastCall;
  if (millis() - lastCall > 250) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 250ms liegt. Wenn nicht, geht der aktuelle Wert verloren.
    lastCall = millis();
    int value; // in cm

    pinMode(ULTRA_SONIC_LEFT_PIN, OUTPUT);
    digitalWrite(ULTRA_SONIC_LEFT_PIN, LOW);
    delayMicroseconds(2);
    digitalWrite(ULTRA_SONIC_LEFT_PIN, HIGH);
    delayMicroseconds(15);
    digitalWrite(ULTRA_SONIC_LEFT_PIN, LOW);
    delayMicroseconds(20);

    pinMode(ULTRA_SONIC_LEFT_PIN, INPUT);
    const unsigned long duration_left = pulseIn(ULTRA_SONIC_LEFT_PIN, HIGH, 29000); // Timeout nach 29ms ca. 5 Meter. Standard-Timeout von 1000ms verzögert den Code stark
    if (duration_left == 0) {
      value = 600;
    } else {
      value = duration_left / 29 / 2; // Umrechnung in cm
    }
    //sendTimestamp();
    sendData(DISTANCE_LEFT, String(value));
    return true;
  } else {
    return false;
  }
}

// misst und sendet die Daten des rechten Ulltraschallsensors
bool sendDistanceRight() {
  static unsigned long lastCall;
  if (millis() - lastCall > 250) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 250ms liegt. Wenn nicht, geht der aktuelle Wert verloren.
    lastCall = millis();
    int value; // in cm

    pinMode(ULTRA_SONIC_RIGHT_PIN, OUTPUT);
    digitalWrite(ULTRA_SONIC_RIGHT_PIN, LOW);
    delayMicroseconds(2);
    digitalWrite(ULTRA_SONIC_RIGHT_PIN, HIGH);
    delayMicroseconds(15);
    digitalWrite(ULTRA_SONIC_RIGHT_PIN, LOW);
    delayMicroseconds(20);

    pinMode(ULTRA_SONIC_RIGHT_PIN, INPUT);
    const unsigned long duration_right = pulseIn(ULTRA_SONIC_RIGHT_PIN, HIGH, 29000);
    if (duration_right == 0) {
      value = 600;
    } else {
      value = duration_right / 29 / 2; // Umrechnung in cm
    }
    //sendTimestamp();
    sendData(DISTANCE_RIGHT, String(value));
    return true;
  } else {
    return false;
  }
}

// misst und sendet die Daten der Phototransistoren
bool sendPTData() {
  static unsigned long lastCall;
  if (millis() - lastCall > 250) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 250ms liegt. Wenn nicht, geht der aktuelle Wert verloren.
    lastCall = millis();
    int ptRightValue = analogRead(PT_RIGHT_PIN);
    int ptLeftValue = analogRead(PT_LEFT_PIN);
    int ptDownValue = analogRead(PT_DOWN_PIN);
    sendTimestamp();
    sendData(PT_RIGHT, String(ptRightValue));
    sendData(PT_LEFT, String(ptLeftValue));
    sendData(PT_DOWN, String(ptDownValue));
    return true;
  } else {
    return false;
  }
}

// Akkustand senden
bool sendBatteryVoltage() {
  static unsigned long lastCall;
  if (millis() - lastCall > 3000) {         // Damit zwischen den einzelnen Aufrufen mindestens ein Abstand von 3000ms liegt. Wenn nicht, geht der aktuelle Wert verloren.
    lastCall = millis();
    int value = analogRead(BATTERY_VOLTAGE_PIN);
    //sendTimestamp();
    sendData(BATTERY_VOLTAGE, String(value));
    return true;
  } else {
    return false;
  }
}

// Zeitstempel senden
void sendTimestamp() {
  sendData(TIMESTAMP, String(millis()));
}

// Set-Up des Electronic Speed Controllers
void setupESC() {
  // Attach the ESC on pin 11
  servoESC.attach(SERVO_ESC_PIN, 1000, 2000);  // (pin, min pulse width, max pulse width in microseconds) 
  
  // Setup ESC to LiPo Mode
  servoESC.write(0);
  delay(2000);              
//  servoESC.write(180);      // full Throttle
//  delay(8000);              // Wait for first *beep*
//  servoESC.write(100);
//  delay(5000);              // wait for final *beeps*
}

void setup() {
  serialObject.begin(9600);
  servoFinUpDown.attach(SERVO_FIN_UP_DOWN_PIN);
  servoFinLeftRight.attach(SERVO_FIN_LEFT_RIGHT_PIN);
  setupESC();
}

void loop() {
  // Sensordaten senden
  sendDistanceLeft();
  sendDistanceRight();
  sendPTData();
  sendBatteryVoltage();

//  unsigned long timestamps[2]; // zur Zeitmessung
//  timestamps[0] = millis();

  // Es wird acht mal hintereinander geprüft, ob weitere serielle Daten vorliegen.
  // Das ist notwendig, da bei jedem Aufruf von serialEventt() die Daten nur bis zum nächsten Zeilenumbruch ausgewertet werden
  // und es durch die Ultraschallsensormessungen sonst zu zu großen Verzögerungen kommen kann.
  for (byte i=0; i<8; i++) {
    if (serialObject.available() > 0) {
      serialEventt();
    }
  }
//  timestamps[1] = millis();
//  sendData(LOG, String(timestamps[1]-timestamps[0]));

  // Servos/ESC werden aktualisiert:
  // wenn der Zeitraum zwischen zwei Aktualisierungen zu kurz ist und der Wert nicht aktualisiert werden kann,
  // wird finUpDownNeedsUpdate auf true gesetzt und im nächsten Durchlauf der Loop-Funktion ein neuer Versuch gestartet
  if (finUpDownNeedsUpdate) {
    finUpDownNeedsUpdate = !writeFinUpDownPosition(newFinUpDownPosition);
    if (!finUpDownNeedsUpdate) {
      sendData(FIN_UP_DOWN, String(newFinUpDownPosition));
    }
  }
  if (finLeftRightNeedsUpdate) {
    finLeftRightNeedsUpdate = !writeFinLeftRightPosition(newFinLeftRightPosition);
    if (!finLeftRightNeedsUpdate) {
      sendData(FIN_LEFT_RIGHT, String(newFinLeftRightPosition));
    }
  }
  if (motorDriveNeedsUpdate) {
    motorDriveNeedsUpdate = !writeMotorDrive(newMotorDrivePosition);
    if (!motorDriveNeedsUpdate) {
      sendData(MOTOR_DRIVE, String(newMotorDrivePosition));
    }
  }
}
