// Klasse zur Nutzung der Daten eines Game-Controllers
public class GamePadControl {
  private ControlIO control;
  private ControlDevice device; // neues Gamepad
  private VirtualController vController;
  private ServoFin finUpDown;
  private ServoFin finLeftRight;
  private ServoFin motorESC;
  private Crossdiagram servoDiagram;
  private Crossdiagram escDiagram;
  
  // benötigte Steuerelemente:
  private ControlButton modeButton;    // zur Aktivierung der Gamepadsteuerung
  private ControlButton AButton;       // um in Mapping-Modus A zu wechseln
  private ControlButton BButton;       // um in Mapping-Modus B zu wechseln
  private ControlButton startButton;   // zum Starten des Motors im MODE_ONE
  
  private ControlSlider sliderLeftX;   // linker Steuerknüppel x-Achse
  private ControlSlider sliderLeftY;   // linker Steuerknüppel y-Achse
  private ControlSlider sliderLeftZ;   // links hinten Aufschrift: LT
  private ControlSlider sliderRightX;  // rechter Steuerknüppel x-Achse
  private ControlSlider sliderRightY;  // rechter Steuerknüppel y-Achse
  private ControlSlider sliderRightZ;  // rechts hinten Aufschrift: RT
  
  private int mappingMode = 0;        // festgelegter Modus zum Mappen der Steuerelemente auf den virtuellen Controller
  public final int MAPPING_A = 0;
  public final int MAPPING_B = 1;
  
  private boolean gamePadControllEnabled = false;  // speichert, ob die Steuerung über den GameController aktiviert/deaktiviert wurde
  
  // Statusse der verwendeten Buttons:
  private boolean modeButtonState = false;
  private boolean modeButtonStateSwitched = false;
  private boolean AButtonState = false;
  private boolean AButtonStateSwitched = false;
  private boolean BButtonState = false;
  private boolean BButtonStateSwitched = false;
  private boolean startButtonState = false;
  private boolean startButtonStateSwitched = false;
  
  // zum Speichern der entsprechenden bisherigen Werte:
  private float motorDrive;
  private float newMotorDrive;
  private float leftRightPosition;
  private float newLeftRightPosition;
  private float upDownPosition;
  private float newUpDownPosition;
  
  
  
  GamePadControl(ControlIO control, VirtualController vController, ServoFin finUpDown, ServoFin finLeftRight, ServoFin motorESC, Crossdiagram servoDiagram, Crossdiagram escDiagram) {
    this.control = control;
    this.vController = vController;
    this.finUpDown = finUpDown;
    this.finLeftRight = finLeftRight;
    this.motorESC = motorESC;
    this.servoDiagram = servoDiagram;
    this.escDiagram = escDiagram;
  }
  
  public void setup() {
    device = control.getDevice("Microsoft X-Box 360 pad");
    if (device == null) {
      println("Game-Controller nicht gefunden");
    } else {
      modeButton = device.getButton("Mode");
      AButton = device.getButton("A");
      BButton = device.getButton("B");
      startButton = device.getButton("Unknown");
      
      sliderLeftX = device.getSlider("x");    // Steuerknüppel links x
      sliderLeftY = device.getSlider("y");    // Steuerknüppel links y
      sliderLeftZ = device.getSlider("z");    // LT (hinten)
      sliderRightX = device.getSlider("rx");  // Steuerknüppel rechts x
      sliderRightY = device.getSlider("ry");  // Steuerknüppel rechts y
      sliderRightZ = device.getSlider("rz");  // RT (hinten)
      
      sliderLeftX.setTolerance(0.2);
      sliderLeftY.setTolerance(0.2);
      sliderRightX.setTolerance(0.2);
      sliderRightY.setTolerance(0.2);
    }
  }
  
  // zum festlegen des Mapping-Modus'
  private void setMappingMode(int mode) {
    mappingMode = mode;
  }
  
  // um die Anzahl der unterschiedlichen Werte zu reduzieren (damit weniger Daten übertragen werden müssen)
  private float adjustValue(float value) {
    if (value <= -0.9) {
      return -1.0;
    } else if (value <= -0.7) {
      return -0.8;
    } else if (value <= -0.5) {
      return -0.5;
    } else if (value <= -0.3) {
      return -0.4;
    } else if (value <= -0.2) {
      return -0.2;
    } else if (value < 0.2) {
      return 0.0;
    } else if (value < 0.3) {
      return 0.2;
    } else if (value < 0.5) {
      return 0.4;
    } else if (value < 0.7) {
      return 0.5;
    } else if (value < 0.9) {
      return 0.8;
    } else if (value >= 0.9){
      return 1.0;
    } else {
      return 0;
    }
  }
  
  // zum Erfassen der aktuellen Zustände und Zustandänderungen der einzelnen Knöpfe:
  
  private void updateModeButtonState() {
    if (modeButton.pressed() != modeButtonState) {
      modeButtonState = !modeButtonState;
      modeButtonStateSwitched = true;
    } else {
      modeButtonStateSwitched = false;
    }
  }
  
  private void updateAButtonState() {
    if (AButton.pressed() != AButtonState) {
      AButtonState = !AButtonState;
      AButtonStateSwitched = true;
    } else {
      AButtonStateSwitched = false;
    }
  }
  
  private void updateBButtonState() {
    if (BButton.pressed() != BButtonState) {
      BButtonState = !BButtonState;
      BButtonStateSwitched = true;
    } else {
      BButtonStateSwitched = false;
    }
  }
  
  private void updateStartButtonState() {
    if (startButton.pressed() != startButtonState) {
      startButtonState = !startButtonState;
      startButtonStateSwitched = true;
    } else {
      startButtonStateSwitched = false;
    }
  }
  
  // Loop-Funktion wird aus der Loop-Funktiion des ManualMode aufgerufen
  public void loop() {
    updateModeButtonState(); // Status des Buttons, mit dem das Gamepad aktiviert werden kann, aktualisieren
    
    if (modeButtonStateSwitched == true && modeButtonState == true) {  // Wenn der modeButton gerade gedrückt wird, wird die Steuerung über das GamePad an oder aus geschaltet
      gamePadControllEnabled = !gamePadControllEnabled;
    }
    
    if (gamePadControllEnabled) {
      // Statusse der Buttons aktualisieren
      updateStartButtonState();
      updateAButtonState();
      updateBButtonState();
      
      // Funktionen der Button bei Bedarf ausführen:
      if (AButtonStateSwitched == true && AButtonState == true) {
        setMappingMode(MAPPING_A);                                  // Mapping A aktivieren
      }
      if (BButtonStateSwitched == true && BButtonState == true) {
        setMappingMode(MAPPING_B);                                  // Mapping B aktivieren
      }
      
      // bisherige Werte des virtuellen Controllers ermitteln:
      motorDrive = vController.getMotorDrive();
      leftRightPosition = vController.getLeftRightPosition();
      upDownPosition = vController.getUpDownPosition();
      
      // je nach Mapping-Modus die neuen Werte des virtuellen Controllers anhand des Gamecontrollers ermitteln:
      switch (mappingMode) {
        case MAPPING_A:
          newMotorDrive = adjustValue((sliderRightZ.getValue()+1)/2);  // motorDrive wird über den rechten hinteren Slider gesteuert
          newLeftRightPosition = adjustValue(sliderLeftX.getValue());  // Drehung um die vertikale Achse erfolgt über den x-Wert des linken Steuerknüppels
          newUpDownPosition = adjustValue(-sliderLeftY.getValue());    // Drehung um die Querachse erfolg über den x-Wert des linken Steuerknüppels
          
          // Wenn sich ein Wert verändert hat, dann: virtuellen Controller aktualisieren, neue Werte an Arduino senden und Diagramm aktualisieren:
          if (motorDrive != newMotorDrive) {
            vController.updateMotorDrive(newMotorDrive);
            motorESC.sendNewPosition((int) map(vController.getMotorDrive(), 0, 1, 0, 180));
            escDiagram.drawFinPosition(map(vController.getMotorDrive(), 0, 1, 0, 180), 90);
          }
          if (leftRightPosition != newLeftRightPosition) {
            vController.updateLeftRightPosition(newLeftRightPosition);
            finLeftRight.sendNewPosition((int) map(vController.getLeftRightPosition(), -1, 1, -45, 45));
            servoDiagram.drawFinPosition(map(vController.getUpDownPosition(), -1, 1, 27, 153), map(vController.getLeftRightPosition(), -1, 1, 45, 135));
          }
          if (upDownPosition != newUpDownPosition) {
            vController.updateUpDownPosition(newUpDownPosition);
            finUpDown.sendNewPosition((int) map(vController.getUpDownPosition(), -1, 1, -63, 63));
            servoDiagram.drawFinPosition(map(vController.getUpDownPosition(), -1, 1, 27, 153), map(vController.getLeftRightPosition(), -1, 1, 45, 135));
          }
          
          break;
        case MAPPING_B:
          // motorDrive als Abstand zwischen Steuerknüppelposition und Nullposition auf einer zweidimensionalen Ebene (Pythagoras)
          newMotorDrive = adjustValue(sqrt((pow(sliderLeftX.getValue(), 2) + pow(sliderLeftY.getValue(), 2))));
          // leftRightPosition als Winkel des Steuerknüppels (Steuerknüppel nach oben (-1, 0) enspricht 0 Grad)
          newLeftRightPosition = adjustValue((sliderLeftX.getValue() < 0 ? -1 : 1) * atan(abs(sliderLeftX.getValue()/sliderLeftY.getValue()))/HALF_PI);
          // upDownPosition als Wert des sliderRightY
          newUpDownPosition = adjustValue(-sliderRightY.getValue());
          
          // Wenn sich ein Wert verändert hat, dann: virtuellen Controller aktualisieren, neue Werte an Arduino senden und Diagramm aktualisieren:
          if (motorDrive != newMotorDrive) {
            vController.updateMotorDrive(newMotorDrive);
            motorESC.sendNewPosition((int) map(vController.getMotorDrive(), 0, 1, 0, 180));
            escDiagram.drawFinPosition(map(vController.getMotorDrive(), 0, 1, 0, 180), 90);
          }
          if (leftRightPosition != newLeftRightPosition) {
            vController.updateLeftRightPosition(newLeftRightPosition);
            println(newLeftRightPosition);
            finLeftRight.sendNewPosition((int) map(vController.getLeftRightPosition(), -1, 1, -45, 45));
            servoDiagram.drawFinPosition(map(vController.getUpDownPosition(), -1, 1, 27, 153), map(vController.getLeftRightPosition(), -1, 1, 45, 135));
          }
          if (upDownPosition != newUpDownPosition) {
            vController.updateUpDownPosition(newUpDownPosition);
            finUpDown.sendNewPosition((int) map(vController.getUpDownPosition(), -1, 1, -63, 63));
            servoDiagram.drawFinPosition(map(vController.getUpDownPosition(), -1, 1, 27, 153), map(vController.getLeftRightPosition(), -1, 1, 45, 135));
          }
          break;
      }
    }
  }
  
}
