public class SerialCommunication {
  private VirtualArduino vArduino;
  private Serial serialObject;
  private Crossdiagram servoDiagram;
  private Crossdiagram escDiagram;
  private LivePlotDiagram PTDiagramVArduino;
  private LivePlotDiagram distanceDiagramVArduino;
  private Battery batteryStatus;
  
  // Keys für die serielle Kommunikation im Format {key:value}[ZEILENUMBRUCH]
  //
  final int TIMESTAMP = 0; // aktueller Zeitstempel
  final int LOG = 250;     // andere Nachrichten
  final int ECHO = 254;    // Echotest
  final int ERROR = 255;   // Für möglicherweise auftretende Fehler (falscher Datentyp, falscher Wertebereich, etc.)
  
  // Codes der Aktuatoren:
  final int FIN_UP_DOWN = 1;
  final int FIN_LEFT_RIGHT = 2;
  final int MOTOR_DRIVE = 3;
  
  // Codes der Sensoren:
  final int DISTANCE_LEFT = 4;
  final int DISTANCE_RIGHT = 5;
  final int PT_RIGHT = 6;
  final int PT_LEFT = 7;
  final int PT_DOWN = 8;

  // Code des Akkus:
  final int BATTERY_VOLTAGE = 9;
  
  
  SerialCommunication(Serial serialObject, VirtualArduino vArduino, Crossdiagram servoDiagram, Crossdiagram escDiagram, LivePlotDiagram PTDiagramVArduino, LivePlotDiagram distanceDiagramVArduino, Battery batteryStatus) {
    this.serialObject = serialObject;
    this.vArduino = vArduino;
    this.servoDiagram = servoDiagram;
    this.escDiagram = escDiagram;
    this.PTDiagramVArduino = PTDiagramVArduino;
    this.distanceDiagramVArduino = distanceDiagramVArduino;
    this.batteryStatus = batteryStatus;
  }
  
  // Sendet ein Key-Value-Pair im Format {key:value}[ZEILENUMBRUCH] über die serielle Schnittstelle
  public boolean sendData(int key, String value) {
    serialObject.write("{" + str(key) + ":" + value + "}\n");
    println("PC: {" + str(key) + ":" + value + "}");
    return true;
  }
  
  // Wird am Ende des Loop-Durchlaufs aufgerufen, wenn seriell empfangene Daten verfügbar sind
  // Sendet Fehlernachricht zurück, wenn Daten unvollständig sind
  // Ruft bei entsprechenden Keys bestimmte Funktionen auf
  public void serialEvent() {
    String data = serialObject.readStringUntil(10);                                    // ein Datenpaket endet jeweileils mit einem Zeilenumbruch
    if (data == null) {
      return;
    } else {
      data = data.replace("\n", "");
    }
    if (!data.startsWith("{") || !data.endsWith("}")) {                                // testet, ob am Anfang und Ende des Datenpaketes jeweils das richtige Zeichen ist '{' bzw. '}'
      final String message = "Anfangs- oder Endtag fehlt. Daten unvollständig.";       // falls nicht -> Fehlermeldung
      println(message);
      return;
    } else {
      int collonPos = data.indexOf(':');
      if (collonPos == -1) {                                                           // testet, ob ein Doppelpunkt als Trenner zwischen Key und Value vorhanden ist
        final String message = "Doppelpunkt als Trenner kann nicht gefunden werden.";  // falls nicht -> Fehlermeldung
        println(message);
        return;
      } else {
        final int key = int(data.substring(1, collonPos));                             // ermittelt den Key
        final String value = data.substring(collonPos + 1, data.length() - 1);         // ermittelt den entsprechenden Wert
        //println(data); // zum Debuggen
        switch (key) {                                                                 // je nach dem welcher Key vorliegt, wird eine der folgenden Aktionen ausgeführt
          case TIMESTAMP:                                                              // im virtuellen Arduino wird der Zeitstempel aktualisiert (wird vor allem zum Diagramm zeichnen verwendet)
            vArduino.updateTimestamp(int(value)/1000.0);                                 // speichert timestamp in Sekunden
            //println(data);
            break;
          case FIN_UP_DOWN:
            vArduino.updateFinUpDownPosition(int(value));                                                            // Wert im virtuellen Arduino aktualisieren
            servoDiagram.drawFinPosition(vArduino.getFinUpDownPosition(), 180-vArduino.getFinLeftRightPosition());   // Diagramm zeichnen
            // println(data);
            break;
          case FIN_LEFT_RIGHT:
            vArduino.updateFinLeftRightPosition(int(value));                                                         // Wert im virtuellen Arduino aktualisieren
            servoDiagram.drawFinPosition(vArduino.getFinUpDownPosition(), vArduino.getFinLeftRightPosition());   // Diagramm zeichnen
            //println(data);
            break;
          case MOTOR_DRIVE:
            vArduino.updateMotorDrivePosition(int(value));                                                           // Wert im virtuellen Arduino aktualisieren
            escDiagram.drawFinPosition(vArduino.getMotorDrivePosition(), 90);                                        // Diagramm zeichnen
            //println(data);
            break;
          case DISTANCE_LEFT:
            vArduino.updateDistanceLeftValue(int(value));                                                            // Wert im virtuellen Arduino aktualisieren
            distanceDiagramVArduino.updateLayer("Distanz links", vArduino.getDistanceLeftValueRingBuffer());         // Diagramm zeichnen
            //println(data);
            break;
          case DISTANCE_RIGHT:
            vArduino.updateDistanceRightValue(int(value));                                                           // Wert im virtuellen Arduino aktualisieren
            distanceDiagramVArduino.updateLayer("Distanz rechts", vArduino.getDistanceRightValueRingBuffer());       // Diagramm zeichnen
            //println(data);
            break;
          case PT_LEFT:
            vArduino.updatePTLeftValue(int(value));                                                                  // Wert im virtuellen Arduino aktualisieren
            PTDiagramVArduino.updateLayer("PT links", vArduino.getPTLeftValueRingBuffer());                          // Diagramm zeichnen
            //println(data);
            break;
          case PT_RIGHT:
            vArduino.updatePTRightValue(int(value));                                                                 // Wert im virtuellen Arduino aktualisieren
            PTDiagramVArduino.updateLayer("PT rechts", vArduino.getPTRightValueRingBuffer());                        // Diagramm zeichnen
            //println(data);
            break;
          case PT_DOWN:
            vArduino.updatePTDownValue(int(value));                                                                  // Wert im virtuellen Arduino aktualisieren
            PTDiagramVArduino.updateLayer("PT unten", vArduino.getPTDownValueRingBuffer());                          // Diagramm zeichnen
            //println(data);
            break;
          case BATTERY_VOLTAGE:
            vArduino.updateBatteryVoltageValue(int(value));                                                          // Wert im virtuellen Arduino aktualisieren
            batteryStatus.drawCharge(map(vArduino.getBatteryVoltageValue(), 0, 1023, 0, 5));                         // Diagramm zeichnen
            println(map(vArduino.getBatteryVoltageValue(), 0, 1023, 0, 5));
            break;
          case LOG:        // Ausgabe des Wertes
            println(data);
            break;
          case ECHO:
            println(data); // Ausgabe des Wertes
            println(millis());
            break;
          case ERROR:      // Ausgabe des Wertes
            println(data);
            break;
          default:
            println("Falscher Key von Arduino gesendet"); // Fehlermeldung
            break;
        }
      }
    }
  }
  
}
